var fs = require('fs');
var path = require('path');
var async = require('async');
var http = require('http');
var https = require('https');
var proxy = require('http-proxy');
var helpers = require('./helpers');
var certificates = require('./certificates');

var app = exports = module.exports = {};

app.init = function(options) {
  this.defaultConfiguration();

  this._target = {
    host: options.target,
    http: true,
    https: true
  };

  this._ports = {
    http: options.http,
    https: options.https
  };

  if(this._ports.https) {
    var opts = this._httpsOptions = {
      key: fs.readFileSync(get_path(options.key)),
      cert: fs.readFileSync(get_path(options.cert)),
      ca: []
    };

    if("string" == typeof options.ca) options.ca = [options.ca];
    options.ca.forEach(function(file) {
      opts.ca.push(fs.readFileSync(get_path(file)));
    });

    opts.SNICallback = this.SNICallback;
    if(options.certs) {
      opts.SNICallback = certificates(options.certs);
    }
  }

  this.use(helpers({ target: this._target }));
};

app.defaultConfiguration = function() {
  this._stack = [];
  this._errStack = [];
  this._proxyServer = proxy.createProxyServer();
};

app.use = function(fn) {
  if(fn.length == 4) return this._errStack.push(fn);

  return this._stack.push(fn);
};

app.route = function(req, res, next) {
  var self = this;
  var stack = this._stack;

  async.eachSeries(stack, function(mw, done) {
    mw.apply(self, [req, res, done]);
  },
  function(err) {
    if(err) {
      self.error.apply(self, [err, req, res, next]);
    }
  });
};

app.error = function(err, req, res, next) {
  var stack = this._errStack;

  async.eachSeries(stack, function(mw, done) {
    mw(err, req, res, done);
  }, function(err) {
    if (err) throw err;
    next(err);
  });
};

app.proxy = function(req, res, next) {
  if(!req.target) return next();
  if(!req.target.https && req.secure) return next();
  if(!req.target.http && !req.secure) return next();

  var self = this;

  if(req.ws) {
    self._proxyServer.ws(req, req._socket, req._head, {
      target: req.target.host,
      xfwd: true,
      agent: http.globalAgent
    }, function(err) {
      self.error.apply(self, [err, req, res, next]);
    });
  } else {
    self._proxyServer.web(req, res, {
      target: req.target.host,
      xfwd: true,
      agent: http.globalAgent
    }, function(err) {
      self.error.apply(self, [err, req, res, next]);
    });
  }
};

app.not_found = function(req, res, next) {
  var err = new Error('Not found');
  err.status = 404;
  
  next(err);
};

app.SNICallback = function(hostname) {
  return null;
};

app.ws = function(req, socket, head) {
  req.ws = true;
  req._socket = socket;
  req._head = head;

  this.route(req, null, function(err) {
    
  });
};

app.listen = function(callback) {
  this._stack.push(this.proxy);
  this._stack.push(this.not_found);

  if(this._ports.http) {
    var httpServer = http.createServer(this);
    httpServer.on('upgrade', this.ws.bind(this));
    httpServer.listen.apply(httpServer, [this._ports.http]);
  }

  if(this._ports.https) {
    var httpsServer = https.createServer(this._httpsOptions, this);
    httpsServer.on('upgrade', this.ws.bind(this));
    httpsServer.listen.apply(httpsServer, [this._ports.https]);
  }

  if(callback) callback();
};

function get_path(filename) {
  var app_path = path.dirname(require.main.filename);

  return path.resolve(app_path, filename);
}